from typing import Callable
from superposedpulses.forcing import PulseParameters

import numpy as np


class TwoPointForcing:
    """Container class with the signal's forcing containing arrival times,
    amplitudes and durations for all the pulses for both signals.

    Random variables for different pulses are independent, but random
    variables for a single pulse may be correlated, even in different
    points.
    """

    def __init__(
        self,
        total_pulses: int,
        arrival_times: np.ndarray,
        amplitudes_a: np.ndarray,
        durations_a: np.ndarray,
        amplitudes_b: np.ndarray,
        durations_b: np.ndarray,
        delays: np.ndarray,
    ):
        assert total_pulses == len(arrival_times)
        if amplitudes_a is not None:
            assert total_pulses == len(amplitudes_a)
        if durations_a is not None:
            assert total_pulses == len(durations_a)
        if amplitudes_b is not None:
            assert total_pulses == len(amplitudes_b)
        if durations_b is not None:
            assert total_pulses == len(durations_b)
        if delays is not None:
            assert total_pulses == len(delays)

        self.total_pulses = total_pulses
        self.arrival_times = arrival_times
        self.amplitudes_a = amplitudes_a
        self.amplitudes_b = amplitudes_b
        self.durations_a = durations_a
        self.durations_b = durations_b
        self.delays = delays

    def get_pulse_parameters_a(self, pulse_index: int) -> PulseParameters:
        return PulseParameters(
            self.arrival_times[pulse_index],
            self.amplitudes_a[pulse_index],
            self.durations_a[pulse_index],
        )

    def get_pulse_parameters_b(self, pulse_index: int) -> PulseParameters:
        return PulseParameters(
            self.arrival_times[pulse_index] + self.delays[pulse_index],
            self.amplitudes_b[pulse_index],
            self.durations_b[pulse_index],
        )


class TwoPointForcingGenerator:
    """Responsible for generating a forcing for a two point model.

    The forcing consists of a set of amplitudes, durations, delays and arrival times.
        amplitudes_a and amplitudes_b are the amplitudes at each point, respectively. By default,
        they are exponentially distributed and uncorrelated.
        durations_a and durations_b are the duration times at each point, respectively. By
        default, they are equal and degenerate distributed.
        delays are the delays of arrival times between point B and point A. By default,
        they are degenerate distributed.
        arrival_times are the arrival times at point A.
    """

    def __init__(self):
        self._amplitude_distribution = lambda k: np.random.default_rng().exponential(
            size=k
        )
        self._duration_distribution = lambda k: np.ones(k)
        self._delay_distribution = lambda k: np.ones(k)

    def get_forcing(self, times: np.ndarray, waiting_time: float) -> TwoPointForcing:
        total_pulses = int(max(times) / waiting_time )
        arrival_times = np.random.default_rng().uniform(
            low=times[0], high=times[len(times) - 1], size=total_pulses
        )
        amplitudes_a = self._amplitude_distribution(total_pulses)
        durations_a = self._duration_distribution(total_pulses)
        amplitudes_b = self._amplitude_distribution(total_pulses)
        durations_b = durations_a
        delays = self._delay_distribution(total_pulses)
        return TwoPointForcing(
            total_pulses,
            arrival_times,
            amplitudes_a,
            durations_a,
            amplitudes_b,
            durations_b,
            delays,
        )

    def set_amplitude_distribution(
        self,
        amplitude_distribution_function: Callable[[int], np.ndarray],
    ):
        self._amplitude_distribution = amplitude_distribution_function

    def set_duration_distribution(
        self, duration_distribution_function: Callable[[int], np.ndarray]
    ):
        self._duration_distribution = duration_distribution_function

    def set_delay_distribution(
        self, delay_distribution_function: Callable[[int], np.ndarray]
    ):
        self._delay_distribution = delay_distribution_function



import matplotlib.pyplot as plt
import numpy as np

import superposedpulses.forcing as frc
import superposedpulses.point_model as pm
import superposedpulses.pulse_shape as ps

# Simplest case, using defaults: exponential pulse shape, exponentially distributed amplitudes, constant duration times.

model = pm.PointModel(waiting_time=10.0, total_duration=100, dt=0.01)
times, signal = model.make_realization()

plt.plot(times, signal)
plt.show()

# Double exponential shape

model = pm.PointModel(waiting_time=10.0, total_duration=100, dt=0.01)
model.set_pulse_shape(ps.StandardPulseGenerator("2-exp", lam=0.35))
times, signal = model.make_realization()

plt.plot(times, signal)
plt.show()

# Adding noise to the signal

model = pm.PointModel(waiting_time=10.0, total_duration=100, dt=0.01)
model.add_noise(noise_to_signal_ratio=0.01, noise_type="additive", seed=None)
times, signal = model.make_realization()

plt.plot(times, signal)
plt.show()

# Say you want to customise your model a bit: use constant amplitude distribution, and box pulse shapes

model = pm.PointModel(waiting_time=10.0, total_duration=100, dt=0.01)
model.set_amplitude_distribution("deg")
model.set_pulse_shape(ps.BoxShortPulseGenerator())

times, signal = model.make_realization()

plt.plot(times, signal)
plt.show()

# If you want to implement your own distributions, you can do so by setting a custom ForcingGenerator. Say you want half
# of your pulses to have amplitude 1, and the other half to have amplitude 2.

model = pm.PointModel(waiting_time=10.0, total_duration=100, dt=0.01)
my_forcing_gen = frc.StandardForcingGenerator()
my_forcing_gen.set_amplitude_distribution(
    lambda k: np.random.randint(low=1, high=3, size=k)
)

model.set_custom_forcing_generator(my_forcing_gen)
times, s = model.make_realization()

plt.plot(times, s)
plt.show()


# Say you want to do something more fancy, for example make a joint distribution of the pulse parameters, then
# you will have to implement your own ForcingGenerator, which is as easy as inheriting from frc.ForcingGenerator and
# set whatever you want in the get_forcing method


class MyFancyForcingGenerator(frc.ForcingGenerator):
    def __init__(self):
        pass

    def get_forcing(self, times: np.ndarray, waiting_time: float) -> frc.Forcing:
        total_pulses = int(max(times) / waiting_time)
        arrival_time_indx = np.random.randint(0, len(times), size=total_pulses)
        amplitudes = np.random.default_rng().exponential(scale=1.0, size=total_pulses)
        durations = amplitudes + np.abs(
            0, np.random.normal(loc=0, scale=0.5, size=len(amplitudes))
        )
        return frc.Forcing(
            total_pulses, times[arrival_time_indx], amplitudes, durations
        )

    def set_amplitude_distribution(
        self,
        amplitude_distribution_function,
    ):
        pass

    def set_duration_distribution(self, duration_distribution_function):
        pass


model = pm.PointModel(waiting_time=0.1, total_duration=1000, dt=0.01)
model.set_custom_forcing_generator(MyFancyForcingGenerator())

times, s = model.make_realization()

plt.plot(times, s)
plt.show()

# You might also want to get the actual pulse parameters that were used in the realization, this can be done by getting
# the forcing, which is the set of amplitudes, arrival times and durations for each pulse.

forcing = model.get_last_used_forcing()
plt.hist(forcing.durations)
plt.show()

# Two point model example
model = pm.TwoPointModel(waiting_time=10.0, total_duration=100, dt=0.01)
model.set_pulse_shape(ps.ExponentialShortPulseGenerator(tolerance=1e-50))
times, signal_a, signal_b = model.make_realization()

plt.plot(times, signal_a)
plt.plot(times, signal_b)
plt.show()



from abc import ABC
from typing import Callable, Tuple, Union
import warnings

import numpy as np
from tqdm import tqdm
from superposedpulses.forcing import (
    Forcing,
    StandardForcingGenerator,
    ForcingGenerator,
    PulseParameters,
)
from superposedpulses.pulse_shape import (
    ShortPulseGenerator,
    ExponentialShortPulseGenerator,
    PulseGenerator,
)
from superposedpulses.two_point_forcing import TwoPointForcingGenerator
from scipy.signal import fftconvolve

__COMMON_DISTRIBUTIONS__ = ["exp", "deg"]


def _get_common_distribution(
    distribution_name: str, average: float
) -> Callable[[int], np.ndarray]:
    if distribution_name == "exp":
        return lambda k: np.random.default_rng().exponential(scale=average, size=k)
    elif distribution_name == "deg":
        return lambda k: average * np.ones(k)
    else:
        raise NotImplementedError


class AbstractModel(ABC):
    """
    Abstract class for FPP Models containing commonly used methods.
    Parameters
    ----------
    waiting_time Average time between pulses
    total_duration Total duration of the process
    dt Time step

    """

    def __init__(self, waiting_time: float, total_duration: float, dt: float):
        self.waiting_time = waiting_time
        self.T = total_duration
        self.dt = dt
        self._times: np.ndarray = np.arange(0, total_duration, dt)
        self._pulse_generator: ShortPulseGenerator = ExponentialShortPulseGenerator()
        self._last_used_forcing = None

    def _add_pulse_to_signal(
        self, signal: np.ndarray, pulse_parameters: PulseParameters
    ):
        """
        Adds a pulse to the provided signal array. Uses self._pulse_generator to generate the pulse shape, this can
        either be a ps.PulseGenerator or a ps.ShortPulseGenerator.
        Parameters
        ----------
        signal Signal array under construction
        pulse_parameters Parameters of the current pulse

        """
        if isinstance(self._pulse_generator, PulseGenerator):
            signal += pulse_parameters.amplitude * self._pulse_generator.get_pulse(
                self._times - pulse_parameters.arrival_time,
                pulse_parameters.duration,
            )
            return

        if isinstance(self._pulse_generator, ShortPulseGenerator):
            cutoff = self._pulse_generator.get_cutoff(pulse_parameters.duration)
            from_index = max(int((pulse_parameters.arrival_time - cutoff) / self.dt), 0)
            to_index = min(
                int((pulse_parameters.arrival_time + cutoff) / self.dt),
                len(self._times),
            )

            pulse = pulse_parameters.amplitude * self._pulse_generator.get_pulse(
                self._times[from_index:to_index] - pulse_parameters.arrival_time,
                pulse_parameters.duration,
            )
            signal[from_index:to_index] += pulse
            return

        raise NotImplementedError(
            "Pulse shape has to inherit from PulseShape or ShortPulseShape"
        )

    def get_last_used_forcing(self):
        """
        Returns the latest used forcing. If several realizations of the process are run only the latest forcing will be
        available.
        -------
        """
        return self._last_used_forcing

    def set_pulse_shape(
        self, pulse_generator: Union[PulseGenerator, ShortPulseGenerator]
    ):
        """
        Parameters
        ----------
        pulse_shape Instance of PulseShape, get_pulse will be called for each pulse when making a realization.
        """
        self._pulse_generator = pulse_generator


class PointModel(AbstractModel):
    """PointModel is a container for all model parameters and is responsible for
    generating a realization of the process through make_realization.

    Uses a ForcingGenerator to generate the forcing, this is by default
    a StandardForcingGenerator.

    Parameters
    ----------
    waiting_time Average time between pulses
    total_duration Total duration of the process
    dt Time step
    """

    def __init__(self, waiting_time: float, total_duration: float, dt: float):
        super(PointModel, self).__init__(waiting_time, total_duration, dt)
        self._forcing_generator: ForcingGenerator = StandardForcingGenerator()
        self._noise = None

    def make_realization(self) -> Tuple[np.ndarray, np.ndarray]:
        result = np.zeros(len(self._times))
        forcing = self._forcing_generator.get_forcing(self._times, waiting_time=self.waiting_time)

        for k in tqdm(range(forcing.total_pulses), position=0, leave=True):
            pulse_parameters = forcing.get_pulse_parameters(k)
            self._add_pulse_to_signal(result, pulse_parameters)

        if self._noise is not None:
            result += self._discretize_noise(forcing)

        self._last_used_forcing = forcing

        return self._times, result

    def set_custom_forcing_generator(self, forcing_generator: ForcingGenerator):
        self._forcing_generator = forcing_generator

    def set_amplitude_distribution(
        self, amplitude_distribution: str, average_amplitude: float = 1.0
    ):
        """Sets the amplitude distribution to be used by the forcing.

        Args:
            amplitude_distribution: str
                'exp': exponential with scale parameter average_amplitude
                'deg': degenerate with location average_amplitude
            average_amplitude: float, defaults to 1.
        """
        if amplitude_distribution in __COMMON_DISTRIBUTIONS__:
            self._forcing_generator.set_amplitude_distribution(
                _get_common_distribution(amplitude_distribution, average_amplitude)
            )
        else:
            raise NotImplementedError

    def set_duration_distribution(
        self, duration_distribution: str, average_duration: float = 1.0
    ):
        """Sets the amplitude distribution to be used by the forcing.

        Args:
            duration_distribution: str
                'exp': exponential with scale parameter average_duration
                'deg': degenerate with location average_duration
            average_duration: float, defaults to 1.
        """
        if duration_distribution in __COMMON_DISTRIBUTIONS__:
            self._forcing_generator.set_duration_distribution(
                _get_common_distribution(duration_distribution, average_duration)
            )
        else:
            raise NotImplementedError

    def add_noise(
        self,
        noise_to_signal_ratio: float,
        seed: Union[None, int] = None,
        noise_type: str = "additive",
    ) -> None:
        """
        Specifies noise for realization.
        Parameters
        ----------
        noise_to_signal_ratio: float, defined as X_rms/S_rms where X is noise and S is signal.
        seed: None or int, seed for the noise generator
        noise_type: str
            "additive": additive noise
            "dynamic": dynamic noise (only applicable for constant duration times)
            "both": both additive and dynamic noise
        """
        assert noise_type in {"additive", "dynamic", "both"}
        assert seed is None or isinstance(seed, int)
        assert noise_to_signal_ratio >= 0

        self._noise_type = noise_type
        self._noise_random_number_generator = np.random.RandomState(seed=seed)

        warnstring = """Calculation of noise rms 
                        (1) assumes average duration time is 1
                        (2) uses numerical mean amplitude"""
        warnings.warn(warnstring)
        mean_amplitude = self._forcing_generator.get_forcing(
            self._times, waiting_time=self.waiting_time
        ).amplitudes.mean()
        gamma = 1./self.waiting_time

        self._sigma = np.sqrt(noise_to_signal_ratio * gamma) * mean_amplitude

        self._noise = np.zeros(len(self._times))

    def _discretize_noise(self, forcing: Forcing) -> np.ndarray:
        """Discretizes noise for the realization"""

        if self._noise_type in {"additive", "both"}:
            self._noise += self._sigma * self._noise_random_number_generator.normal(
                size=len(self._times)
            )

        if self._noise_type in {"dynamic", "both"}:
            durations = forcing.durations
            pulse_duration_constant = np.all(durations == durations[0])
            assert (
                pulse_duration_constant
            ), "Dynamic noise is only applicable for constant duration times."

            kern = self._pulse_generator.get_pulse(
                np.arange(-self._times[-1] / 2, self._times[-1] / 2, self.dt),
                durations[0],
            )
            dW = self._noise_random_number_generator.normal(
                scale=np.sqrt(2 * self.dt), size=len(self._times)
            )
            self._noise += self._sigma * fftconvolve(dW, kern, "same")

        return self._noise


class TwoPointModel(AbstractModel):
    def __init__(self, waiting_time: float, total_duration: float, dt: float):
        super(TwoPointModel, self).__init__(waiting_time, total_duration, dt)
        self._forcing_generator: TwoPointForcingGenerator = TwoPointForcingGenerator()

    def make_realization(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        signal_a = np.zeros(len(self._times))
        signal_b = np.zeros(len(self._times))
        forcing = self._forcing_generator.get_forcing(self._times, waiting_time=self.waiting_time)

        for k in tqdm(range(forcing.total_pulses), position=0, leave=True):
            pulse_parameters_a = forcing.get_pulse_parameters_a(k)
            pulse_parameters_b = forcing.get_pulse_parameters_b(k)
            self._add_pulse_to_signal(signal_a, pulse_parameters_a)
            self._add_pulse_to_signal(signal_b, pulse_parameters_b)

        self._last_used_forcing = forcing
        return self._times, signal_a, signal_b

    def set_custom_forcing_generator(self, forcing_generator: TwoPointForcingGenerator):
        self._forcing_generator = forcing_generator



from abc import ABC, abstractmethod
from typing import Callable

import numpy as np


class PulseParameters:
    """Container class with the parameters (arrival time, amplitude and
    duration) for a single pulse."""

    def __init__(self, arrival_time: float, amplitude: float, duration: float):
        self.arrival_time = arrival_time
        self.amplitude = amplitude
        self.duration = duration


class Forcing:
    """Container class with the signal forcing containing arrival times,
    amplitudes and durations for all the pulses."""

    def __init__(
        self,
        total_pulses: int,
        arrival_times: np.ndarray,
        amplitudes: np.ndarray,
        durations: np.ndarray,
    ):
        assert total_pulses == len(arrival_times)
        if amplitudes is not None:
            assert total_pulses == len(amplitudes)
        if durations is not None:
            assert total_pulses == len(durations)

        self.total_pulses = total_pulses
        self.arrival_times = arrival_times
        self.amplitudes = amplitudes
        self.durations = durations

    def get_pulse_parameters(self, pulse_index: int) -> PulseParameters:
        return PulseParameters(
            self.arrival_times[pulse_index],
            self.amplitudes[pulse_index],
            self.durations[pulse_index],
        )


class ForcingGenerator(ABC):
    """Abstract class used by PointModel to generate forcing.

    Implementations of this class should have a get_forcing method,
    returning a forcing with arrival times, amplitudes and durations for
    all pulses.
    """

    @abstractmethod
    def get_forcing(self, times: np.ndarray, waiting_time: float) -> Forcing:
        raise NotImplementedError

    @abstractmethod
    def set_amplitude_distribution(
        self,
        amplitude_distribution_function: Callable[[int], np.ndarray],
    ):
        raise NotImplementedError

    @abstractmethod
    def set_duration_distribution(
        self, duration_distribution_function: Callable[[int], np.ndarray]
    ):
        raise NotImplementedError


class StandardForcingGenerator(ForcingGenerator):
    """Generates a standard forcing, with uniformly distributed arrival times.

    The resulting process is therefore a Poisson process. Amplitude and
    duration distributions can be customized.
    """

    def __init__(self):
        self._amplitude_distribution = None
        self._duration_distribution = None

    def get_forcing(self, times: np.ndarray, waiting_time: float) -> Forcing:
        total_pulses = int(max(times) / waiting_time)
        arrival_times = np.random.default_rng().uniform(
            low=times[0], high=times[len(times) - 1], size=total_pulses
        )
        amplitudes = self._get_amplitudes(total_pulses)
        durations = self._get_durations(total_pulses)
        return Forcing(total_pulses, arrival_times, amplitudes, durations)

    def set_amplitude_distribution(
        self,
        amplitude_distribution_function: Callable[[int], np.ndarray],
    ):
        self._amplitude_distribution = amplitude_distribution_function

    def set_duration_distribution(
        self, duration_distribution_function: Callable[[int], np.ndarray]
    ):
        self._duration_distribution = duration_distribution_function

    def _get_amplitudes(self, total_pulses) -> np.ndarray:
        if self._amplitude_distribution is not None:
            return self._amplitude_distribution(total_pulses)
        return np.random.default_rng().exponential(scale=1.0, size=total_pulses)

    def _get_durations(self, total_pulses) -> np.ndarray:
        if self._duration_distribution is not None:
            return self._duration_distribution(total_pulses)
        return np.ones(total_pulses)



from abc import ABC, abstractmethod
import numpy as np
from typing import Callable
import warnings


class PulseGenerator(ABC):
    """Abstract pulse shape, implementations should return a pulse shape vector
    given an array of time stamps with (possibly) different durations.

    For long time series this becomes expensive, and a ShortPulseShape
    implementation is preferred if the pulse function decays fast to 0
    or under machine error.
    """

    @abstractmethod
    def get_pulse(self, times: np.ndarray, duration: float) -> np.ndarray:
        """
        Abstract method, implementations should return np.ndarray with pulse shape with the same length as times.
        Parameters
        ----------
        times time array under which the pulse function is to be evaluated
        duration Duration time

        Returns
        -------
        np.ndarray with the pulse shape,

        """
        raise NotImplementedError


class StandardPulseGenerator(PulseGenerator):
    """Generates all pulse shapes previously supported."""

    __SHAPE_NAMES__ = {"1-exp", "lorentz", "2-exp", "gaussian"}
    # TODO: Implement the others

    def __init__(self, shape_name: str = "1-exp", **kwargs):
        """
        Parameters
        ----------
        shape_name Should be one of StandardPulseShapeGenerator.__SHAPE_NAMES__
        kwargs Additional arguments to be passed to special shapes:
            - "2-exp":
            - "lam" parameter for the asymmetry parameter
        """
        assert (
            shape_name in StandardPulseGenerator.__SHAPE_NAMES__
        ), "Invalid shape_name"
        self._shape_name: str = shape_name
        self._kwargs = kwargs

    def get_pulse(
        self, times: np.ndarray, duration: float, tolerance: float = 1e-5
    ) -> np.ndarray:

        kern = self._get_generator(self._shape_name)(times, duration, self._kwargs)
        err = max(np.abs(kern[0]), np.abs(kern[-1]))
        if err > tolerance:
            warnings.warn("Value at end point of kernel > tol, end effects may occur.")
        return kern

    @staticmethod
    def _get_generator(
        shape_name: str,
    ) -> Callable[[np.ndarray, float, dict], np.ndarray]:
        if shape_name == "1-exp":
            return StandardPulseGenerator._get_exponential_shape
        if shape_name == "2-exp":
            return StandardPulseGenerator._get_double_exponential_shape
        if shape_name == "lorentz":
            return StandardPulseGenerator._get_lorentz_shape
        if shape_name == "gaussian":
            return StandardPulseGenerator._get_gaussian_shape

    @staticmethod
    def _get_exponential_shape(
        times: np.ndarray, duration: float, kwargs
    ) -> np.ndarray:
        kern = np.zeros(len(times))
        kern[times >= 0] = np.exp(-times[times >= 0] / duration)
        return kern

    @staticmethod
    def _get_lorentz_shape(times: np.ndarray, duration: float, kwargs) -> np.ndarray:
        return (np.pi * (1 + (times / duration) ** 2)) ** (-1)

    @staticmethod
    def _get_double_exponential_shape(
        times: np.ndarray, duration: float, kwargs
    ) -> np.ndarray:
        lam = kwargs["lam"]
        assert (lam > 0.0) & (lam < 1.0)
        kern = np.zeros(len(times))
        kern[times < 0] = np.exp(times[times < 0] / lam / duration)
        kern[times >= 0] = np.exp(-times[times >= 0] / (1 - lam) / duration)
        return kern

    @staticmethod
    def _get_gaussian_shape(times: np.ndarray, duration: float, kwargs) -> np.ndarray:
        return np.exp(-((times / duration) ** 2) / 2) / np.sqrt(2 * np.pi)


class ShortPulseGenerator(ABC):
    """Abstract pulse shape, implementations should return a pulse shape vector
    with (possibly) different durations.

    The length of the returned array is not restricted, this is useful
    for pulse shapes such as the exponential or the box pulse, for which
    the signal becomes zero or under machine error very quickly.

    Implementations are responsible of deciding where to place the cutoff for the returned array.
    """

    def __init__(self, tolerance: float = 1e-50):
        self.tolerance = tolerance

    @abstractmethod
    def get_pulse(self, times: np.ndarray, duration: float) -> np.ndarray:
        """
        Abstract method, implementations should return np.ndarray with pulse shape. The returned array should contain
        the values of the shape for the times [-T, T] with sampling dt, where T is such that the value of the pulse
        function is under self.tolerance. The center of the pulse should be located at the center of the array.
        Parameters
        ----------
        dt Sampling time
        duration Duration time

        Returns
        -------
        np.ndarray with the pulse shape,

        """
        raise NotImplementedError

    @abstractmethod
    def get_cutoff(self, duration: float) -> float:
        """
        Abstract method, implementations should return the cutoff above (and below) which the pulse becomes negligible,
        i.e. return T such that abs(p(T/duration)) < self.tolerance and abs(p(-T/duration)) < self.tolerance
        Parameters
        ----------
        duration duration of the pulse

        Returns
        -------
        Cutoff to cut the pulse
        """
        raise NotImplementedError


class ExponentialShortPulseGenerator(ShortPulseGenerator):
    def __init__(
        self, lam: float = 0, tolerance: float = 1e-50, max_cutoff: float = 1e50
    ):
        """Exponential pulse generator, the length of the returned array is
        dynamically set to be the shortest to reach a pulse value under the
        given tolerance. That is, if the pulse shape is p(t), the returned
        array will be p(t) with t in [-T, T] such that p(-T), p(T) < tolerance.

        If a lam argument is provided different than 0, the pulse shape will be
        a double exponential:

        p(t) = exp(t/lam) for t<0; p(t) = exp(-t/(1-lam)) for t>=0.

        A max_cutoff is provided to avoid returning pulse arrays of arbitrarily long lengths.
        Parameters
        ----------
        tolerance Maximum error when cutting the pulse.
        max_cutoff
        lam Asymmetry parameter, defaults to 0 (exponential pulse shape).
        """
        super(ExponentialShortPulseGenerator, self).__init__(tolerance)
        self._max_cutoff = max_cutoff
        self.lam = lam

    def get_pulse(self, times: np.ndarray, duration: float) -> np.ndarray:
        kern = np.zeros(len(times))
        if self.lam == 0:
            kern[times >= 0] = np.exp(-times[times >= 0] / duration)
            return kern
        kern[times >= 0] = np.exp(-times[times >= 0] / (duration * (1 - self.lam)))
        kern[times < 0] = np.exp(times[times < 0] / (duration * self.lam))
        return kern

    def get_cutoff(self, duration: float) -> float:
        cutoff = -duration * np.log(self.tolerance)
        return min(cutoff, self._max_cutoff)


class LorentzShortPulseGenerator(ShortPulseGenerator):
    def __init__(self, tolerance: float = 1e-50, max_cutoff: float = 1e50):
        """Lorentz pulse generator, the length of the returned array is
        dynamically set to be the shortest to reach a pulse value under the
        given tolerance. That is, if the pulse shape is p(t), the returned
        array will be p(t) with t in [-T, T] such that p(-T), p(T) < tolerance.

        p(t) = 1/ (pi (t^2 + 1)).

        A max_cutoff is provided to avoid returning pulse arrays of arbitrarily long lengths.
        Parameters
        ----------
        tolerance Maximum error when cutting the pulse.
        max_cutoff
        """
        super(LorentzShortPulseGenerator, self).__init__(tolerance)
        self._max_cutoff = max_cutoff

    def get_pulse(self, times: np.ndarray, duration: float) -> np.ndarray:
        kern = np.zeros(len(times))
        kern = (np.pi * (1 + (times / duration) ** 2)) ** (-1)
        return kern

    def get_cutoff(self, duration: float) -> float:
        cutoff = duration * np.sqrt(1.0 / (self.tolerance * np.pi) - 1)
        return min(cutoff, self._max_cutoff)


class GaussianShortPulseGenerator(ShortPulseGenerator):
    def __init__(self, tolerance: float = 1e-50, max_cutoff: float = 1e50):
        """
        Gaussian pulse generator. The length of the returned array is
        dynamically set to be the shortest to reach a pulse value under the
        given tolerance. That is, if the pulse shape is p(t), the returned
        array will be p(t) with t in [-T, T] such that p(-T), p(T) < tolerance.

        p(t) = exp(-(t)^2 / 2) / sqrt(2*pi).

        A max_cutoff is provided to avoid returning pulse arrays of arbitrarily long lengths.
        Parameters
        ----------
        tolerance Maximum error when cutting the pulse.
        max_cutoff

        """
        super(GaussianShortPulseGenerator, self).__init__(tolerance)
        self._max_cutoff = max_cutoff

    def get_pulse(self, times: np.ndarray, duration: float) -> np.ndarray:
        kern = np.zeros(len(times))
        kern = np.exp(-((times / duration) ** 2) / 2) / np.sqrt(2 * np.pi)
        return kern

    def get_cutoff(self, duration: float) -> float:
        cutoff = duration * np.sqrt(-2 * np.log(np.sqrt(2 * np.pi) * self.tolerance))
        return min(cutoff, self._max_cutoff)


class BoxShortPulseGenerator(ShortPulseGenerator):
    """Box shape p(t, tau):

    p(t, tau) = 1, for |t| < tau / 2
    p(t, tau) = 0, otherwise
    """

    def __init__(self, tolerance: float = 1e-50):
        super(BoxShortPulseGenerator, self).__init__(tolerance)

    def get_pulse(self, times: np.ndarray, duration: float) -> np.ndarray:
        kern = np.zeros(len(times))
        kern[abs(times) < duration / 2] = 1
        return kern

    def get_cutoff(self, duration: float) -> float:
        return duration


class TriangularShortPulseGenerator(ShortPulseGenerator):
    """Triangular shape p(t, tau):

    p(t, tau) = 1 - |t|/tau, for |t| < tau
    p(t, tau) = 0, otherwise
    """

    def __init__(self, tolerance: float = 1e-50):
        super(TriangularShortPulseGenerator, self).__init__(tolerance)

    def get_pulse(self, times: np.ndarray, duration: float) -> np.ndarray:
        kern = np.zeros(len(times))
        kern[abs(times) < duration] = 1 - abs(times[abs(times) < duration]) / duration
        return kern

    def get_cutoff(self, duration: float) -> float:
        return duration



import os
from setuptools import setup

name = "superposed-pulses"

with open("README.md") as f:
    long_description = f.read()

here = os.path.abspath(os.path.dirname(__file__))

setup(
    name=name,
    description="Python scripts used by the fusion energy group at UiT The Arctic University of Norway.",
    author="Juan Manuel Losada",
    author_email="juan.m.losada@uit.no",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/uit-cosmo/filtered-point-process",
    download_url="https://github.com/uit-cosmo/3ppy/archive/refs/tags/1.1.tar.gz",
    license="MiT",
    version="1.4",
    packages=["superposedpulses"],
    python_requires=">=3.0",
    install_requires=[
        "numpy>=1.15.0",
        "scipy>=1.4.0",
        "tqdm>=4.50.2",
    ],
    classifiers=[
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    zip_safe=False,
)
